/**
 MPI    MPI,   ,    MPI.
  :
MPI_Init(),
MPI_Finalize(),
MPI_Send(),
MPI_Recv(),
MPI_Isend(),
MPI_Irecv(),
MPI_Issend(),
MPI_Barrier(),
MPI_Abort(),
MPI_Comm_rank(),
MPI_Comm_size(),
MPI_Get_count(),
MPI_Test().
\begin{verbatim}
   ,   ,    .
      . 
        MPI.
      mpi.h
-------------------------------------------------------------------------------------------------------------
   :
1)   mpiemul.lib
2)      mpi.h,    
3)   NUMBER_OF_PROCESS.      .
-------------------------------------------------------------------------------------------------------------
,          -      Microsoft Visual Studio 6,    .
@memo  MPI
@name  Mpiemul
@author (c) 2004 A. . 
@version 1.1  \today
*/
#define NUMBER_OF_PROCESS 3
#define MPI_IDENT     0
#define MPI_CONGRUENT 1
#define MPI_SIMILAR   2
#define MPI_UNEQUAL   3

typedef int MPI_Datatype;
#define MPI_CHAR           ((MPI_Datatype)1)
#define MPI_UNSIGNED_CHAR  ((MPI_Datatype)2)
#define MPI_BYTE           ((MPI_Datatype)3)
#define MPI_SHORT          ((MPI_Datatype)4)
#define MPI_UNSIGNED_SHORT ((MPI_Datatype)5)
#define MPI_INT            ((MPI_Datatype)6)
#define MPI_UNSIGNED       ((MPI_Datatype)7)
#define MPI_LONG           ((MPI_Datatype)8)
#define MPI_UNSIGNED_LONG  ((MPI_Datatype)9)
#define MPI_FLOAT          ((MPI_Datatype)10)
#define MPI_DOUBLE         ((MPI_Datatype)11)
#define MPI_LONG_DOUBLE    ((MPI_Datatype)12)
#define MPI_LONG_LONG_INT  ((MPI_Datatype)13)

#define MPI_PACKED         ((MPI_Datatype)14)
#define MPI_LB             ((MPI_Datatype)15)
#define MPI_UB             ((MPI_Datatype)16)

#define MPI_FLOAT_INT      ((MPI_Datatype)17)
#define MPI_DOUBLE_INT     ((MPI_Datatype)18)
#define MPI_LONG_INT       ((MPI_Datatype)19)
#define MPI_SHORT_INT      ((MPI_Datatype)20)
#define MPI_2INT           ((MPI_Datatype)21)
#define MPI_LONG_DOUBLE_INT ((MPI_Datatype)22)

#define MPI_COMPLEX        ((MPI_Datatype)23)
#define MPI_DOUBLE_COMPLEX ((MPI_Datatype)24)
#define MPI_LOGICAL        ((MPI_Datatype)25)
#define MPI_REAL           ((MPI_Datatype)26)
#define MPI_DOUBLE_PRECISION ((MPI_Datatype)27)
#define MPI_INTEGER        ((MPI_Datatype)28)
#define MPI_2INTEGER       ((MPI_Datatype)29)
#define MPI_2COMPLEX       ((MPI_Datatype)30)
#define MPI_2DOUBLE_COMPLEX   ((MPI_Datatype)31)
#define MPI_2REAL             ((MPI_Datatype)32)
#define MPI_2DOUBLE_PRECISION ((MPI_Datatype)33)
#define MPI_CHARACTER         ((MPI_Datatype)1)

typedef int MPI_Comm;
#define MPI_COMM_WORLD 91
#define MPI_COMM_SELF 92

typedef int MPI_Group;
#define MPI_GROUP_EMPTY 90

typedef int MPI_Op;

#define MPI_MAX    (MPI_Op)(100)
#define MPI_MIN    (MPI_Op)(101)
#define MPI_SUM    (MPI_Op)(102)
#define MPI_PROD   (MPI_Op)(103)
#define MPI_LAND   (MPI_Op)(104)
#define MPI_BAND   (MPI_Op)(105)
#define MPI_LOR    (MPI_Op)(106)
#define MPI_BOR    (MPI_Op)(107)
#define MPI_LXOR   (MPI_Op)(108)
#define MPI_BXOR   (MPI_Op)(109)
#define MPI_MINLOC (MPI_Op)(110)
#define MPI_MAXLOC (MPI_Op)(111)

#define MPI_TAG_UB 81
#define MPI_HOST 83
#define MPI_IO 85
#define MPI_WTIME_IS_GLOBAL 87

#define MPIR_TAG_UB 80
#define MPIR_HOST 82
#define MPIR_IO 84
#define MPIR_WTIME_IS_GLOBAL 86

#define MPI_COMM_NULL      ((MPI_Comm)0)
#define MPI_OP_NULL        ((MPI_Op)0)
#define MPI_GROUP_NULL     ((MPI_Group)0)
#define MPI_DATATYPE_NULL  ((MPI_Datatype)0)
#define MPI_REQUEST_NULL   ((MPI_Request)0)
#define MPI_ERRHANDLER_NULL ((MPI_Errhandler )0)

#define MPI_MAX_PROCESSOR_NAME 256
#define MPI_MAX_ERROR_STRING   512
#define MPI_MAX_NAME_STRING     63	

#define MPI_UNDEFINED      (-32766)
#define MPI_UNDEFINED_RANK MPI_UNDEFINED
#define MPI_KEYVAL_INVALID 0

/* Upper bound on the overhead in bsend for each message buffer */
#define MPI_BSEND_OVERHEAD 512

/* Topology types */
#define MPI_GRAPH  1
#define MPI_CART   2

#define MPI_BOTTOM      (void *)0

#define MPI_PROC_NULL   (-1)
#define MPI_ANY_SOURCE 	(-2)
#define MPI_ANY_TAG	(-1)

typedef struct { 
    int count;
    int MPI_SOURCE;
    int MPI_TAG;
    int MPI_ERROR;
} MPI_Status;

#if (defined(_SX) && !defined(_LONG64))

typedef long long MPI_Aint;
#else
typedef long MPI_Aint;
#endif

#if defined(USE_STDARG) 
typedef void (MPI_Handler_function) MPIR_ARGS(( MPI_Comm *, int *, ... ));
#else
typedef void (MPI_Handler_function) ();
#endif

#define MPI_ERRORS_ARE_FATAL ((MPI_Errhandler)119)
#define MPI_ERRORS_RETURN    ((MPI_Errhandler)120)
#define MPIR_ERRORS_WARN     ((MPI_Errhandler)121)
typedef int MPI_Errhandler;
#define MPI_NULL_COPY_FN   MPIR_null_copy_fn
#define MPI_NULL_DELETE_FN MPIR_null_delete_fn
#define MPI_DUP_FN         MPIR_dup_fn

typedef int *MPI_Request;

#define MPI_VERSION 1
#define MPI_SUBVERSION 1

#define MPICH_HAS_HANDLE2INT

typedef void *MPI_Handle_type;

typedef int  MPI_Fint;
typedef enum { MPI_DATATYPE_HANDLE, MPI_REQUEST_HANDLE, MPI_COMM_HANDLE, 
	       MPI_GROUP_HANDLE, MPI_ERRHANDLE_HANDLE, MPI_OP_HANDLE } 
MPI_Handle_enum;

#include "mpi_errno.h"

#ifdef TRAPPER
#define TRAPPER_ANIMATION
#endif

#include "windows.h"

#ifdef TRAPPER
#include "trapperb.h"
#endif

#ifdef _WINDOWS
extern int wp4_putchar(int);
extern int wp4_fputchar(int);
extern int wp4_vprintf (const char *fmt, va_list ap);
extern int wp4_printf (const char *fmt, ...);
extern int wp4_scanf (const char *format, ...);
#endif
#ifndef MPI

extern int NumberOfProcesses(int size);
extern int MPI_Init(int *argc,char ***argv);
extern int MPI_Finalize();
extern int MPI_Send(void	*buf, int count,  \
											   MPI_Datatype datatype, \
											   int dest, int msgtag,  \
											   MPI_Comm comm);
extern int MPI_Recv(void* buf, int count,   \
											  MPI_Datatype datatype,  \
											  int source, int msgtag, \
											  MPI_Comm comm,          \
											  MPI_Status *status);
extern int MPI_Comm_size( MPI_Comm comm, int* size);
extern int MPI_Comm_rank( MPI_Comm comm, int* rank);
extern int MPI_Barrier( MPI_Comm comm);
extern int MPI_Abort( MPI_Comm comm, int errorcode);
extern int MPI_Isend(void* buf, int count,  \
											   MPI_Datatype datatype, \
											   int dest, int msgtag,  \
											   MPI_Comm comm,         \
											   MPI_Request *request);
extern int MPI_Issend(void* buf, int count,  \
											   MPI_Datatype datatype, \
											   int dest, int msgtag,  \
											   MPI_Comm comm,         \
											   MPI_Request *request);

extern int MPI_Irecv(void *buf, int count,  \
											   MPI_Datatype datatype, \
											   int source, int msgtag,\
											   MPI_Comm comm,         \
											   MPI_Request *request);
extern int MPI_Get_count( MPI_Status *status, MPI_Datatype datatype, int *count);
extern int MPI_Test( MPI_Request *request, int *flag, MPI_Status *status);

#ifdef NUMBER_OF_PROCESS
#define MPI_Init(argc, argv) NumberOfProcesses(NUMBER_OF_PROCESS);MPI_Init(argc, argv)
#endif
#endif

#ifdef MPI
extern "C"  __declspec(dllexport)int NumberOfProcesses(int size);
extern "C"  __declspec(dllexport)int MPI_Init(int *argc,char ***argv);
extern "C"  __declspec(dllexport)int MPI_Finalize();
extern "C"  __declspec(dllexport)int MPI_Send(void	*buf, int count,  \
											   MPI_Datatype datatype, \
											   int dest, int msgtag,  \
											   MPI_Comm comm);
extern "C"  __declspec(dllexport)int MPI_Recv(void* buf, int count,   \
											  MPI_Datatype datatype,  \
											  int source, int msgtag, \
											  MPI_Comm comm,          \
											  MPI_Status *status);
extern "C"  __declspec(dllexport)int MPI_Comm_size( MPI_Comm comm, int* size);
extern "C"  __declspec(dllexport)int MPI_Comm_rank( MPI_Comm comm, int* rank);
extern "C"  __declspec(dllexport)int MPI_Barrier( MPI_Comm comm);
extern "C"  __declspec(dllexport)int MPI_Abort( MPI_Comm comm, int errorcode);
extern "C"  __declspec(dllexport)int MPI_Isend(void* buf, int count,  \
											   MPI_Datatype datatype, \
											   int dest, int msgtag,  \
											   MPI_Comm comm,         \
											   MPI_Request *request);

extern "C"  __declspec(dllexport)int MPI_Irecv(void *buf, int count,  \
											   MPI_Datatype datatype, \
											   int source, int msgtag,\
											   MPI_Comm comm,         \
											   MPI_Request *request);
extern "C"  __declspec(dllexport)int MPI_Get_count( MPI_Status *status, MPI_Datatype datatype, int *count);
extern "C"  __declspec(dllexport)int MPI_Test( MPI_Request *request, int *flag, MPI_Status *status);

struct TQueue {
	void *buf;
	int count;
	int IsEmpty;
	MPI_Datatype datatype;
	int dest;
	int source;
	int msgtag;
	MPI_Comm comm;
	TQueue *Next;
	TQueue *Head;
	MPI_Request request;
};
#endif